@@ -73,6 +73,7 @@
 #define PKG_DUMP_LICENSE		(((uint64_t) 1) << 45)
 #define PKG_SOLUTION			(((uint64_t) 1) << 46)
 #define PKG_EXISTS_CFLAGS		(((uint64_t) 1) << 47)
+#define PKG_FRAGMENT_TREE		(((uint64_t) 1) << 48)
 
 static pkgconf_client_t pkg_client;
 static const pkgconf_fragment_render_ops_t *want_render_ops = NULL;
@@ -334,7 +335,7 @@ print_solution_node(pkgconf_client_t *client, pkgconf_pkg_t *pkg, void *unused)
 	(void) client;
 	(void) unused;
 
-	printf("%s (%"PRIu64")\n", pkg->id, pkg->identifier);
+	printf("%s (%"PRIu64")%s\n", pkg->id, pkg->identifier, (pkg->flags & PKGCONF_PKG_PROPF_VISITED_PRIVATE) == PKGCONF_PKG_PROPF_VISITED_PRIVATE ? " [private]" : "");
 }
 
 static bool
@@ -763,6 +764,49 @@ apply_simulate(pkgconf_client_t *client, pkgconf_pkg_t *world, void *data, int m
 }
 #endif
 
+static void
+print_fragment_tree_branch(pkgconf_list_t *fragment_list, int indent)
+{
+	pkgconf_node_t *iter;
+
+	PKGCONF_FOREACH_LIST_ENTRY(fragment_list->head, iter)
+	{
+		pkgconf_fragment_t *frag = iter->data;
+
+		if (frag->type)
+			printf("%*s'-%c%s' [type %c]\n", indent, "", frag->type, frag->data, frag->type);
+		else
+			printf("%*s'%s' [untyped]\n", indent, "", frag->data);
+
+		print_fragment_tree_branch(&frag->children, indent + 2);
+	}
+
+	if (fragment_list->head != NULL)
+		printf("\n");
+}
+
+static bool
+apply_fragment_tree(pkgconf_client_t *client, pkgconf_pkg_t *world, void *data, int maxdepth)
+{
+	pkgconf_list_t unfiltered_list = PKGCONF_LIST_INITIALIZER;
+	int eflag;
+
+	(void) data;
+
+	eflag = pkgconf_pkg_cflags(client, world, &unfiltered_list, maxdepth);
+	if (eflag != PKGCONF_PKG_ERRF_OK)
+		return false;
+
+	eflag = pkgconf_pkg_libs(client, world, &unfiltered_list, maxdepth);
+	if (eflag != PKGCONF_PKG_ERRF_OK)
+		return false;
+
+	print_fragment_tree_branch(&unfiltered_list, 0);
+	pkgconf_fragment_free(&unfiltered_list);
+
+	return true;
+}
+
 static void
 print_license(pkgconf_client_t *client, pkgconf_pkg_t *pkg, void *data)
 {
@@ -903,6 +947,7 @@ usage(void)
 #endif
 	printf("  --fragment-filter=types           filter output fragments to the specified types\n");
 	printf("  --env=prefix                      print output as shell-compatible environmental variables\n");
+	printf("  --fragment-tree                   visualize printed CFLAGS/LIBS fragments as a tree\n");
 
 	printf("\nreport bugs to <%s>.\n", PACKAGE_BUGREPORT);
 }
@@ -1093,6 +1138,7 @@ main(int argc, char *argv[])
 		{ "license", no_argument, &want_flags, PKG_DUMP_LICENSE },
 		{ "verbose", no_argument, NULL, 55 },
 		{ "exists-cflags", no_argument, &want_flags, PKG_EXISTS_CFLAGS },
+		{ "fragment-tree", no_argument, &want_flags, PKG_FRAGMENT_TREE },
 		{ NULL, 0, NULL, 0 }
 	};
 
@@ -1298,24 +1344,25 @@ main(int argc, char *argv[])
 	if ((want_flags & PKG_INTERNAL_CFLAGS) == PKG_INTERNAL_CFLAGS)
 		want_client_flags |= PKGCONF_PKG_PKGF_DONT_FILTER_INTERNAL_CFLAGS;
 
-	/* if these selectors are used, it means that we are inquiring about a single package.
-	 * so signal to libpkgconf that we do not want to use the dependency resolver for more than one level,
-	 * and also limit the SAT problem to a single package.
+	/* if these selectors are used, it means that we are querying metadata.
+	 * so signal to libpkgconf that we only want to walk the flattened dependency set.
 	 */
-	if (((want_flags & PKG_REQUIRES) == PKG_REQUIRES ||
-		(want_flags & PKG_REQUIRES_PRIVATE) == PKG_REQUIRES_PRIVATE ||
-		(want_flags & PKG_PROVIDES) == PKG_PROVIDES ||
-		(want_flags & PKG_VARIABLES) == PKG_VARIABLES ||
-		(want_flags & PKG_PATH) == PKG_PATH ||
-		want_variable != NULL))
-	{
-		maximum_package_count = 1;
+	if ((want_flags & PKG_MODVERSION) == PKG_MODVERSION ||
+	    (want_flags & PKG_REQUIRES) == PKG_REQUIRES ||
+	    (want_flags & PKG_REQUIRES_PRIVATE) == PKG_REQUIRES_PRIVATE ||
+	    (want_flags & PKG_PROVIDES) == PKG_PROVIDES ||
+	    (want_flags & PKG_VARIABLES) == PKG_VARIABLES ||
+	    (want_flags & PKG_PATH) == PKG_PATH ||
+	    want_variable != NULL)
 		maximum_traverse_depth = 1;
-	}
 
-	/* we also want to walk only the flattened dependencies if we are requesting --modversion. */
-	if ((want_flags & PKG_MODVERSION) == PKG_MODVERSION)
-		maximum_traverse_depth = 1;
+	/* if we are asking for a variable, path or list of variables, this only makes sense
+	 * for a single package.
+	 */
+	if ((want_flags & PKG_VARIABLES) == PKG_VARIABLES ||
+	    (want_flags & PKG_PATH) == PKG_PATH ||
+	    want_variable != NULL)
+		maximum_package_count = 1;
 
 	if (getenv("PKG_CONFIG_ALLOW_SYSTEM_CFLAGS") != NULL)
 		want_flags |= PKG_KEEP_SYSTEM_CFLAGS;
@@ -1665,6 +1712,13 @@ cleanup3:
 		apply_requires_private(&pkg_client, &world, NULL, 2);
 	}
 
+	if ((want_flags & PKG_FRAGMENT_TREE))
+	{
+		want_flags &= ~(PKG_CFLAGS|PKG_LIBS);
+
+		apply_fragment_tree(&pkg_client, &world, NULL, 2);
+	}
+
 	if ((want_flags & PKG_CFLAGS))
 	{
 		apply_cflags(&pkg_client, &world, NULL, 2);
